import numpy as np
# for testing against R
from rpy2.robjects import r
import rpy2.robjects.numpy2ri
rpy2.robjects.numpy2ri.activate()

from pyhmc import integrated_autocorr2
from pyhmc.hmc import _check_random_state

def generate_AR1(phi, sigma, n_steps, c=0, y0=0, random_state=None):
    y = np.zeros(n_steps)
    y[0] = y0
    random = _check_random_state(random_state)
    rand = random.normal(scale=sigma, size=(n_steps,))
    for i in range(1, n_steps):
        y[i] = c + phi*y[i-1] + rand[i]
    return y


def test_1():
    r("require('coda')")

    random = np.random.RandomState(1)
    for i in range(10):
        x = generate_AR1(phi=0.95, sigma=1, n_steps=1000, c=0, y0=0, random_state=random)
        r.assign('x', x)
        tau = r('nrow(x)/effectiveSize(x)')[0]
        np.testing.assert_approx_equal(tau, integrated_autocorr2(x))

def test_2():
    x = np.random.randn(100, 2)
    integrated_autocorr2(x)



from __future__ import division
import numpy as np
from pyhmc.autocorr4 import integrated_autocorr4
from pyhmc.tests.test_autocorr2 import generate_AR1

# for testing against R
from rpy2.robjects import r
import rpy2.robjects.numpy2ri
rpy2.robjects.numpy2ri.activate()


def integrated_autocorr4r(x):
    r.require('coda')
    # r.assign('x', x)
    # print(r('spectrum0(x, max.freq=0.1)'))
    # print(r('spectrum0(x, max.freq=0.5)'))
    # print(r('spectrum0(x, max.freq=0.5)'))
    value = r['spectrum0'](x)[0][0] / np.var(x, ddof=1)
    return value


def test_1():
    for i in range(10):
        y = generate_AR1(phi=0.98, sigma=1, n_steps=10000, c=0, y0=0)
        val = integrated_autocorr4(y)
        rval = integrated_autocorr4r(y)
        np.testing.assert_almost_equal(val, rval, decimal=2)


def test_2():
    y = generate_AR1(phi=0.98, sigma=1, n_steps=1000, c=0, y0=0)
    y2 = np.vstack((y, y)).T
    val = integrated_autocorr4(y2)
    print(val)



from __future__ import division, absolute_import, print_function
import numpy as np


def integrated_autocorr5(x, size='cbrt'):
    r"""Estimate the integrated autocorrelation time, :math:`\tau_{int}` of a
    time series.

    This method uses the consistent non-overlapping batch means estimator [1]
    where the number of batches are chosen as functions of the overall run
    length.

    Parameters
    ----------
    x : ndarray, shape=(n_samples, n_dims)
        The time series, with time along axis 0.
    size : {'sqrt', 'cbrt'}
        The batch size. The default value is "sqroot", which uses the square
        root of the sample size. "cuberoot" will cause the function to use the
        cube root of the sample size. A numeric value may be provided if
        neither "sqrt" nor "cbrt" is satisfactory.

    References
    ----------
    .. [1] Flegal, J. M., Haran, M. and Jones, G. L. (2008) Markov chain Monte
       Carlo: Can we trust the third significant figure? Statistical Science,
       23, 250-260.

    Returns
    -------
    tau_int : ndarray, shape=(n_dims,)
      The estimated integrated autocorrelation time of each dimension in
      ``x``, considered independently.
    """
    if x.ndim == 1:
        x = x.reshape(-1, 1)

    if size == 'sqrt':
        batch_size = int(np.sqrt(x.shape[0]))
    elif size == 'cbrt':
        batch_size = int(x.shape[0]**0.5)
    elif np.isscalar(size):
        batch_size = size
    else:
        raise NotImplementedError('unrecoginized argument, size=%s' % size)

    bigvar = np.var(x, axis=0, ddof=1)
    # leave off the extra bit at the end that's not a clean multiple
    x = x[:batch_size*(len(x)//batch_size)]
    bigmean = np.mean(x, axis=0)

    sigma2_bm = np.zeros(x.shape[1])
    for j in range(x.shape[1]):
        # reshape into the batches, and then compute the batch-means
        bm = x[:, j].reshape(-1, batch_size).mean(axis=1)
        sigma2_bm[j] = (batch_size / (len(bm)-1)) * np.sum((bm-bigmean[j])**2)

    return sigma2_bm / bigvar




from __future__ import division, absolute_import, print_function
import numpy as np
from statsmodels.tsa.stattools import acf
from pyhmc._utils import find_first


def integrated_autocorr6(x, c=6):
    r"""Estimate the integrated autocorrelation time, :math:`\tau_{int}` of a
    time series.

    This method performancs a summation of empirical autocorrelation function,
    using Sokal's "automatic windowing" procedure. The window length, ``M`` is
    chosen self-consistently to be the smallest value such that ``M`` is at
    least ``c`` times the estimated autocorrelation time, where ``c`` should
    be a constant in the range of 4, 6, or 10. See Appendix C of Sokal 1988.

    Parameters
    ----------
    x : ndarray, shape=(n_samples, n_dims)
        The time series, with time along axis 0.
    max_length : int
        The data ``x`` is aggregated if necessary by taking batch means so that
        the length of the series is less than ``max.length``.

    References
    ----------
    .. [1] Madras, Neal, and Alan D. Sokal. "The pivot algorithm: a highly
       efficient Monte Carlo method for the self-avoiding walk." J.
       Stat. Phys. 50.1-2 (1988): 109-186.

    Returns
    -------
    tau_int : ndarray, shape=(n_dims,)
        The estimated integrated autocorrelation time of each dimension in
        ``x``, considered independently.
    """
    if x.ndim == 1:
        x = x.reshape(-1, 1)

    tau = np.zeros(x.shape[1])
    for j in range(x.shape[1]):
        f = acf(x[:,j], nlags=len(x), unbiased=False, fft=True)
        # vector of the taus, all with different choices of the window
        # length
        taus = 1 + 2*np.cumsum(f)[1:]
        ms = np.arange(len(f)-1)
        ind = find_first((ms > c*taus).astype(np.uint8))
        tau[j] = taus[ind]
    return tau




#     # reshape and thens sum over the second axis to get the sum of the pairs
    #     # [1,2,3,4,5,6,7,8] -> [[1,2], [3,4], [5,6], [7,8]] -> [3, 7, 11, 15]
    #     gamma = f.reshape(-1, 2).sum(axis=1)
    #     ind = find_first((gamma<0).astype(np.uint8))
    #     tau[j] = -1 + 2*np.sum(gamma[:ind])
    # return tau



from __future__ import division, absolute_import, print_function
import numpy as np
from statsmodels.tsa.stattools import acf
from pyhmc._utils import find_first


def integrated_autocorr3(x):
    r"""Estimate the integrated autocorrelation time, :math:`\tau_{int}` of a
    time series.

    This method performancs a summation of empirical autocorrelation function,
    setting the window based on the initial sequence estimator (Geyer 1992),
    which stops when the sum of two consecutive elements in the empirica
    autocorrelation function become negative.

    Parameters
    ----------
    x : ndarray, shape=(n_samples, n_dims)
        The time series, with time along axis 0.
    max_length : int
        The data ``x`` is aggregated if necessary by taking batch means so that
        the length of the series is less than ``max.length``.


    References
    ----------
    .. [1] Geyer, Charles J. "Practical markov chain monte carlo." Statistical
       Science (1992): 473-483.

    Returns
    -------
    tau_int : ndarray, shape=(n_dims,)
        The estimated integrated autocorrelation time of each dimension in
        ``x``, considered independently.
    """
    if x.ndim == 1:
        x = x.reshape(-1, 1)

    tau = np.zeros(x.shape[1])
    for j in range(x.shape[1]):
        f = acf(x[:,j], nlags=2*(len(x)//2)-1, unbiased=False, fft=True)
        # reshape and thens sum over the second axis to get the sum of the pairs
        # [1,2,3,4,5,6,7,8] -> [[1,2], [3,4], [5,6], [7,8]] -> [3, 7, 11, 15]
        gamma = f.reshape(-1, 2).sum(axis=1)
        ind = find_first((gamma<0).astype(np.uint8))
        tau[j] = -1 + 2*np.sum(gamma[:ind])
    return tau



import numpy as np
from pyhmc._utils import find_first
from scipy.optimize import curve_fit
from pyhmc.tests.test_autocorr2 import generate_AR1
from pyhmc import hmc, integrated_autocorr1

SAMPLES = None
PHI = 0.95

def setup():
    global SAMPLES
    SAMPLES = generate_AR1(PHI, 1, 10000, random_state=0).reshape(-1,1)


def test_2():
    corr_time1 = integrated_autocorr1(SAMPLES)

    # http://www.hep.fsu.edu/~berg/teach/mcmc08/material/lecture07mcmc3.pdf
    # For a large exponential autocorrelation time t_exp, the approximation
    # is that the integrated autocorrelation time is equal to twice the
    # exponential autocorrelation time, which for a AR1 model is 1/log(\phi)

    expected = -2/np.log(PHI)
    assert corr_time1.shape == (1,)
    assert np.abs(corr_time1 - expected) < 0.3 * expected


def test_find_first():
    X = np.array([
         [0, 0, 1, 1, 1],
         [0, 0, 1, 0, 0],
         [0, 1, 0, 0, 0],
         [0, 0, 0, 0, 0],
         [1, 1, 1, 1, 1]], dtype=np.uint8).T

    reference = [2, 2, 1, -1, 0]
    for j in range(X.shape[1]):
        np.testing.assert_array_equal(
            find_first(X[:,j]), reference[j])


def test_3():
    samples = np.hstack((SAMPLES, SAMPLES))
    corr_time = integrated_autocorr1(samples)
    assert corr_time.shape == (2, )



from __future__ import division, absolute_import, print_function
import numpy as np
from scipy.linalg import toeplitz



def integrated_autocorr2(x):
    r"""Estimate the integrated autocorrelation time, :math:`\tau_{int}` of a
    time series.

    This method estimates the spectral density at zero frequency by fitting
    an AR(p) model, with p selected by AIC.

    Parameters
    ----------
    x : ndarray, shape=(n_samples, n_dims)
        The time series, with time along axis 0.

    References
    ----------
    .. [1] Plummer, M., Best, N., Cowles, K., and Vines, K. (2006). CODA:
        Convergence diagnosis and output analysis for MCMC. R News, 6(1):7-11.

    Returns
    -------
    tau_int : ndarray, shape=(n_dims,)
        The estimated integrated autocorrelation time of each dimension in
        ``x``, considered independently.
    """
    if x.ndim == 1:
        x = x.reshape(-1, 1)
    process_var = np.var(x, axis=0, ddof=1)

    tau_int = np.zeros(x.shape[1])
    for j in range(x.shape[1]):
        # fit an AR(p) model, with p selected by AIC
        rho, sigma2 = yule_walker(x[:,j], order_max=10)
        # power spectral density at zero frequency
        spec0 = sigma2 / (1 - np.sum(rho))**2
        # divide by the variance
        tau_int[j] = spec0 / process_var[j]

    return tau_int


def yule_walker(X, aic=True, order_max=None, demean=True):
    """Estimate AR(p) parameters from a sequence X using Yule-Walker equation.

    Parameters
    ----------
    X : array-like
        1d array
    aic: bool
        If ``True``, then the Akaike Information Criterion is used to choose
        the order of the autoregressive model. If ``False``, the model of order
        ``order.max`` is fitted.
    order_max : integer, optional
        Maximum order of model to fit. Defaults to the smaller of N-1 and
        10*log10(N) where N is the length of the sequence.
    demean : bool
        True, the mean is subtracted from `X` before estimation.

    Returns
    -------
    rho : array, shape=(order,)
        The autoregressive coefficients
    sigma2 : float
        Variance of the nosie term
    aic : float
        Akaike Information Criterion
    """
    # this code is adapted from https://github.com/statsmodels/statsmodels.
    # changes are made to increase compability with R's ``ar.yw``.
    X = np.array(X)
    if demean:
        X -= X.mean()
    n = X.shape[0]

    if X.ndim > 1 and X.shape[1] != 1:
        raise ValueError("expecting a vector to estimate AR parameters")

    if order_max is None:
        order_max = min(n - 1, int(10 * np.log10(n)))

    r = np.zeros(order_max+1, np.float64)
    r[0] = (X**2).sum() / n

    for k in range(1, order_max+1):
        r[k] = (X[0:-k]*X[k:]).sum() / n

    orders = np.arange(1, order_max+1) if aic else [order_max]
    aics = np.zeros(len(orders))
    sigmasqs = np.zeros(len(orders))
    rhos = [None for i in orders]

    for i, order in enumerate(orders):
        r_left = r[:order]
        r_right = r[1:order+1]

        # R = toeplitz(r[:-1])
        R = toeplitz(r_left)
        # rho = np.linalg.solve(R, r[1:])
        rho = np.linalg.solve(R, r_right)
        # sigmasq = r[0] - (r[1:]*rho).sum()
        sigmasq = r[0] - (r_right*rho).sum()
        aic = len(X) * (np.log(sigmasq) + 1) + 2*order + 2*demean
        # R compability
        sigmasq = sigmasq * len(X)/(len(X) - (order + 1))

        aics[i] = aic
        sigmasqs[i] = sigmasq
        rhos[i] = rho

    index = np.argmin(aics)
    return rhos[index], sigmasqs[index]



import numpy as np
from pyhmc.tests.test_autocorr2 import generate_AR1
from pyhmc import integrated_autocorr5

# for testing against R
from rpy2.robjects import r
import rpy2.robjects.numpy2ri
rpy2.robjects.numpy2ri.activate()


def test_1():
    x = generate_AR1(0.95, 1, 50**2+25, random_state=0)
    val = integrated_autocorr5(x, size='sqrt')[0]

    r.require('batchmeans')
    r.assign('x', x)
    ref = r('(bm(x)$se)^2 * length(x) / var(x)')[0]
    np.testing.assert_almost_equal(val, ref)


def test_2():
    x = generate_AR1(0.95, 1, 50**2+25, random_state=0)
    x2 = np.vstack((x, x)).T
    val = integrated_autocorr5(x2, size='sqrt')
    assert val.shape == (2,)
    assert val[0] == val[1]



import numpy as np
from pyhmc.tests.test_autocorr2 import generate_AR1
from pyhmc import integrated_autocorr6, integrated_autocorr1


def test_1():
    x = generate_AR1(0.95, 1, 50**2+25, random_state=0)
    x2 = np.vstack((x, x)).T

    t1 = integrated_autocorr6(x)
    t2 = integrated_autocorr6(x2)
    assert t2.shape == (2,)
    assert t2[0] == t1 and t2[1] == t1[0]



# This code is adapted from https://github.com/dfm/emcee (MIT license)
from __future__ import division, print_function, absolute_import
import sys
import numpy as np
from ._utils import find_first
from statsmodels.tsa.stattools import acf

__all__ = ["integrated_autocorr1"]


def integrated_autocorr1(x, acf_cutoff=0.0):
    r"""Estimate the integrated autocorrelation time, :math:`\tau_{int}` of a
    time series.

    This method performancs a summation of empirical autocorrelation function,
    using a window length, ``M``, to the smallest value such that
    ``ACF(m) <= acf_cutoff``. This procedure is used in (Chodera 2007) with
    ``acf_cutoff = 0``. In (Hoffman 2011, Hub 2010), this estimator is used
    with ``acf_cutoff = 0.05``.

    Parameters
    ----------
    x : ndarray, shape=(n_samples, n_dims)
        The time series, with time along axis 0.

    References
    ----------
    .. [1] J. D. Chodera, W. C. Swope, J. W. Pitera, C. Seok, and K. A. Dill.
       JCTC 3(1):26-41, 2007.
    .. [2] Hoffman, M. D., and A. Gelman. "The No-U-Turn sampler: Adaptively
       setting path lengths in Hamiltonian Monte Carlo." arXiv preprint
       arXiv:1111.4246 (2011).
    .. [3] Hub, J. S., B. L. De Groot, and D. V. Der Spoel. "g_wham: A Fre
       Weighted Histogram Analysis Implementation Including Robust Error and
       Autocorrelation Estimates." J. Chem. Theory Comput. 6.12 (2010):
       3713-3720.

    Returns
    -------
    tau_int : ndarray, shape=(n_dims,)
        The estimated integrated autocorrelation time of each dimension in
        ``x``, considered independently.
    """
    # Compute the autocorrelation function.
    if x.ndim == 1:
        x = x.reshape(-1, 1)
    n = len(x)

    tau = np.zeros(x.shape[1])
    for j in range(x.shape[1]):
        f = acf(x[:,j], nlags=n, unbiased=False, fft=True)
        window = find_first((f <= acf_cutoff).astype(np.uint8))
        tau[j] = 1 + 2*f[1:window].sum()

    return tau



import numpy as np
from pyhmc import integrated_autocorr3
from pyhmc.tests.test_autocorr2 import generate_AR1

# for testing against R
from rpy2.robjects import r
import rpy2.robjects.numpy2ri
rpy2.robjects.numpy2ri.activate()


def test_1():
    y = generate_AR1(0.95, 1, 10000)
    tau = integrated_autocorr3(y)

    r.require('mcmc')
    r.assign('x', y)
    r('popvar = (var(x)*(nrow(x)-1)/nrow(x))')
    r('init = initseq(x)')
    tau_ref = r('initseq(x)$var.pos / popvar')[0]
    print(tau, tau_ref)
    np.testing.assert_array_almost_equal(tau, tau_ref)


def test_2():
    y = generate_AR1(0.95, 1, 10000)
    y2 = np.vstack((y, y)).T
    tau = integrated_autocorr3(y2)
    assert tau.shape == (2,)
    assert tau[0] == tau[1]



from __future__ import division, absolute_import, print_function
import numpy as np
import statsmodels.api as sm


def integrated_autocorr4(x, max_length=200):
    r"""Estimate the integrated autocorrelation time, :math:`\tau_{int}` of a
    time series.

    This method fits a linear regression to the lower frequency of the
    log spectrum, which is extrapolated to zero to estimate the zero
    frequency power spectrum.

    Parameters
    ----------
    x : ndarray, shape=(n_samples, n_dims)
        The time series, with time along axis 0.
    max_length : int
        The data ``x`` is aggregated if necessary by taking batch means so that
        the length of the series is less than ``max.length``.

    References
    ----------
    .. [1] Heidelberger, P., and P. D. Welch. "A spectral method for
       confidence interval generation and run length control in simulations."
       Communications of the ACM 24.4 (1981): 233-245.

    Returns
    -------
    tau_int : ndarray, shape=(n_dims,)
        The estimated integrated autocorrelation time of each dimension in
        ``x``, considered independently.
    """
    if x.ndim == 1:
        x = x.reshape(-1, 1)

    pvar = np.var(x, axis=0, ddof=1)

    v0 = np.zeros(x.shape[1])
    for j in range(x.shape[1]):
        xx = x[:, j]
        batch_size = 1
        if max_length is not None and len(x) > max_length:
            batch_size = int(len(x) / max_length)
            xx = xx[:max_length*batch_size].reshape(max_length, batch_size).mean(axis=1)

        v0[j] = _spectrum0(xx) * batch_size
    return v0 / pvar


def _spectrum0(x):
    """The spectral density at frequency zero is estimated by fitting a
    glm to the low-frequency end of the periodogram.

    The raw periodogram is calculated for the series 'x' and a
    generalized linear model with family 'Gamma' and log link is
    fitted to the periodogram. The predictor is proportional to frequency.

    Notes
    -----
    This code is translated from R's CODA package (do.spectrum0).
    """
    N = len(x)
    Nfreq = int(N/2)
    freq = 1 / N + np.arange(Nfreq) / N

    xfft = np.fft.fft(x)
    spec = ((xfft * np.conjugate(xfft)).real / N)[1:Nfreq+1]

    one = np.ones(Nfreq),
    f1 = np.sqrt(3) * (4 * freq - 1),
    df = np.vstack((one, f1)).T

    model = sm.GLM(spec, df, sm.families.Gamma(sm.families.links.log)).fit()
    v0 = model.predict([1, -np.sqrt(3)])[0]
    return v0



